@@ -20,10 +20,9 @@ NAR.fit(epochs=200,
 
 loss_history =  NAR.model.return_loss()
 
-# evaluate the inferred noise model for a given input range
-phi_range = np.linspace(-20, 20,1000)
-noise_model = NAR.na_noisemodel(sequences,
-                                input_range=phi_range)
+# evaluate the inferred noise model
+#noise_model = NAR.na_noisemodel(sequences)
+noise_model, phi_range, latent_trait = NAR.na_noisemodel(sequences, gauge_fix=True)
 
 # plot results using helper function
 na_plots_for_mavenn_demo(loss_history, NAR, noise_model, phi_range)
