@@ -3,8 +3,11 @@ from mavenn.src.error_handling import handle_errors, check
 from mavenn.src.utils import onehot_encode_array, \
     _generate_nbr_features_from_sequences, _generate_all_pair_features_from_sequences
 
+from mavenn.src.utils import _center_matrix, fix_gauge_neighbor_model, fix_gauge_pairwise_model
+
 import numpy as np
 import tensorflow as tf
+import pandas as pd
 
 import tensorflow.keras
 from tensorflow.keras.models import Sequential, Model
@@ -96,7 +99,7 @@ class GlobalEpistasisModel:
 
         self.x_train, self.y_train = self.X, self.y
 
-        print('One-hot encoding...')
+        #print('One-hot encoding...')
         # # onehot-encode sequences
         # self.input_seqs_ohe = []
         # for _ in range(len(self.x_train)):
@@ -149,6 +152,9 @@ class GlobalEpistasisModel:
         check(isinstance(self.y, (list, np.ndarray)),
               'type(y) = %s must be of type list or np.array' % type(self.y))
 
+        check(len(self.X) == len(self.y),
+              'length of inputs (X, y) must be equal')
+
         # check that model type valid
         check(self.model_type in {'additive', 'neighbor', 'pairwise'},
               'model_type = %s; must be "additive", "neighbor", or "pairwise"' %
@@ -395,10 +401,58 @@ class GlobalEpistasisModel:
 
         return self.history
 
+    def latent_model_parameters(self):
+
+        """
+        method that returns a flattened list of gauge-fixed
+        GE latent trait model parameters.
+
+        returns
+        -------
+        theta_gf: (array-like)
+            gauge fixed latent trait models
+        """
+
+        # use training sequences to compute phi
+        get_1st_layer_output = K.function([self.model.layers[0].input], [self.model.layers[1].output])
+
+        phi_for_computing_norm = get_1st_layer_output([self.input_seqs_ohe])
+
+        # compute the norm using phi_for_computing_norm as:
+        norm = np.sqrt(np.var(phi_for_computing_norm[0]))
+
+        # variables required for gauge fixed latent trait parameters theta
+        sequence_length = len(self.x_train[0])
+        # non-gauge fixed theta
+        theta = self.model.layers[1].get_weights()[0]
+
+        # use gauge-fixing method for the appropriate latent model type
+        if self.model_type == 'additive':
+            theta_df = pd.DataFrame(theta.reshape(sequence_length, len(self.bases)), columns=self.bases)
+            # compute mean centered, thus gauge-fixed, additive latent model
+            gf_theta, theta_bar = _center_matrix(theta_df)
+            gf_theta = gf_theta.values
+            # compute offset term which separates latent trait from gauge fixed latent trait
+            a = np.sum(self.input_seqs_ohe[0].reshape(sequence_length, len(self.bases)) * theta_bar)
+            a = -a
+
+        elif self.model_type == 'neighbor':
+
+            # compute gauge-fixed, neighbor latent model
+            gf_theta, a = fix_gauge_neighbor_model(sequence_length, len(self.bases), theta)
+
+        elif self.model_type == 'pairwise':
+
+            # compute gauge-fixed pairwise latent model
+            gf_theta, a = fix_gauge_pairwise_model(sequence_length, len(self.bases), theta)
+
+        return gf_theta.ravel() / norm
+
     def ge_nonlinearity(self,
                         sequences,
                         input_range=None,
-                        gauge_fix=True):
+                        gauge_fix = True,
+                        ):
 
         """
         Method used to plot GE non-linearity.
@@ -414,9 +468,6 @@ class GlobalEpistasisModel:
             If this is none than range will be determined from min
             and max of the latent trait
 
-        gauge_fix: (bool)
-            if true parameters used to compute latent trait will be
-            gauge fixed
 
         returns
         -------
@@ -424,7 +475,6 @@ class GlobalEpistasisModel:
             the nonlinear GE function.
 
         """
-
         # TODO input checks on input_range
 
         # need to do check if linear model, i.e. when no GE network,
@@ -432,7 +482,6 @@ class GlobalEpistasisModel:
 
         # one hot encode sequences and then subsequently use them
         # to compute latent trait
-
         if self.model_type=='additive':
             # one-hot encode sequences in batches in a vectorized way
             sequences_ohe = onehot_encode_array(sequences, self.bases)
@@ -463,23 +512,80 @@ class GlobalEpistasisModel:
             if len(layer.get_weights()[0]) == 1:
                 break
 
-
         for layer in self.model.layers[-latent_trait_layer_index:]:
-        #for layer in self.model.layers[-2:]:
+        # for layer in self.model.layers[-2:]:
             next_input = layer(next_input)
 
+        # the following model is not gauge fixed
         ge_model = Model(inputs=ge_model_input, outputs=next_input)
 
-        if input_range is None:
-            input_range = np.linspace(min(latent_trait), max(latent_trait), 1000)
-            ge_nonlinearity = ge_model.predict(input_range)
+        # variables required for gauge fixed latent trait parameters theta
+        sequence_length = len(self.x_train[0])
+        # non-gauge fixed theta
+        theta = self.model.layers[1].get_weights()[0]
+        bias = self.model.layers[1].get_weights()[1]
 
-            # if input range not provided, return input range
-            # so ge nonliearity can be plotted against it.
-            return ge_nonlinearity, input_range, latent_trait
-        else:
-            ge_nonlinearity = ge_model.predict(input_range)
-            return ge_nonlinearity
+        # use gauge-fixing method for the appropriate latent model type
+        if self.model_type == 'additive':
+            theta_df = pd.DataFrame(theta.reshape(sequence_length, len(self.bases)), columns=self.bases)
+            # compute mean centered, thus gauge-fixed, additive latent model
+            gf_theta, theta_bar = _center_matrix(theta_df)
+            gf_theta = gf_theta.values
+            # compute offset term which separates latent trait from gauge fixed latent trait
+            a = np.sum(sequences_ohe[0].reshape(sequence_length, len(self.bases)) * theta_bar)
+            a = -a
+
+        elif self.model_type == 'neighbor':
+
+            # compute gauge-fixed, neighbor latent model
+            gf_theta, a = fix_gauge_neighbor_model(sequence_length, len(self.bases), theta)
+
+        elif self.model_type == 'pairwise':
+
+            # compute gauge-fixed pairwise latent model
+            gf_theta, a = fix_gauge_pairwise_model(sequence_length, len(self.bases), theta)
+
+        # this conditional currently outputs the not gauge-fixed, original, GE nonlinearity
+        if gauge_fix is False:
+            if input_range is None:
+                input_range = np.linspace(min(latent_trait), max(latent_trait), 1000)
+                ge_nonlinearity = ge_model.predict(input_range)
+
+                # if input range not provided, return input range
+                # so ge nonliearity can be plotted against it.
+                return ge_nonlinearity, input_range, latent_trait
+            else:
+                ge_nonlinearity = ge_model.predict(input_range)
+                return ge_nonlinearity
+
+        # this part of the condition returns the gauge fixed latent trait
+        elif gauge_fix is True:
+
+            # first compute norm which will be used to rescale the parameters theta
+            # compute phi using training sequences
+            phi_for_computing_norm = get_1st_layer_output([self.input_seqs_ohe])
+
+            # compute the norm using phi_for_computing_norm as:
+            norm = np.sqrt(np.var(phi_for_computing_norm[0]))
+
+            # rescale the gf_theta
+            theta_gf_rescaled = gf_theta.ravel() / norm
+
+            # create a copy of the model which will gauge fixed, this
+            # way the user can retain non-gauge-fixed model if they want
+            gauge_fixed_model = tf.keras.models.clone_model(self.model)
+
+            # update the weights of the latent model with the gauge fixed parameters, and compute gauge fixed phi
+            gauge_fixed_model.layers[1].set_weights([theta_gf_rescaled.reshape(len(theta_gf_rescaled), 1),
+                                              self.model.layers[1].get_weights()[1]])
+
+            get_gauge_fixed_phi = K.function([gauge_fixed_model.layers[0].input], [gauge_fixed_model.layers[1].output])
+            phi_gauge_fixed = get_gauge_fixed_phi([sequences_ohe])
+
+            input_range_gf_phi = np.linspace(min(phi_gauge_fixed[0]), max(phi_gauge_fixed[0]), 1000)
+
+            ge_nonlinearity_gf = ge_model.predict([(input_range_gf_phi-bias)*norm+a+bias])
+            return ge_nonlinearity_gf, input_range_gf_phi, phi_gauge_fixed[0]
 
 
 class NoiseAgnosticModel:
@@ -550,7 +656,7 @@ class NoiseAgnosticModel:
 
         self.x_train, self.y_train = self.X, self.y
 
-        print('One-hot encoding...')
+        #print('One-hot encoding...')
         # # onehot-encode sequences
         # self.input_seqs_ohe = []
         # for _ in range(len(self.x_train)):
@@ -594,7 +700,19 @@ class NoiseAgnosticModel:
         """
         Validate parameters passed to the NoiseAgnosticRegression constructor
         """
-        pass
+        check(isinstance(self.X, (list, np.ndarray)),
+              'type(X) = %s must be of type list or np.array' % type(self.X))
+
+        check(isinstance(self.y, (list, np.ndarray)),
+              'type(y) = %s must be of type list or np.array' % type(self.y))
+
+        check(len(self.X) == len(self.y),
+              'length of inputs (X, y) must be equal')
+
+        # check that model type valid
+        check(self.model_type in {'additive', 'neighbor', 'pairwise'},
+              'model_type = %s; must be "additive", "neighbor", or "pairwise"' %
+              self.model_type)
 
     def define_model(self,
                      monotonic=True,
@@ -790,6 +908,54 @@ class NoiseAgnosticModel:
 
         return self.model
 
+
+    def latent_model_parameters(self):
+
+        """
+        method that returns a flattened list of gauge-fixed
+        NA latent trait model parameters.
+
+        returns
+        -------
+        theta_gf: (array-like)
+            gauge fixed latent trait models
+        """
+
+        # use training sequences to compute phi
+        get_1st_layer_output = K.function([self.model.layers[0].input], [self.model.layers[1].output])
+
+        phi_for_computing_norm = get_1st_layer_output([self.input_seqs_ohe])
+
+        # compute the norm using phi_for_computing_norm as:
+        norm = np.sqrt(np.var(phi_for_computing_norm[0]))
+
+        # variables required for gauge fixed latent trait parameters theta
+        sequence_length = len(self.x_train[0])
+        # non-gauge fixed theta
+        theta = self.model.layers[1].get_weights()[0]
+
+        # use gauge-fixing method for the appropriate latent model type
+        if self.model_type == 'additive':
+            theta_df = pd.DataFrame(theta.reshape(sequence_length, len(self.bases)), columns=self.bases)
+            # compute mean centered, thus gauge-fixed, additive latent model
+            gf_theta, theta_bar = _center_matrix(theta_df)
+            gf_theta = gf_theta.values
+            # compute offset term which separates latent trait from gauge fixed latent trait
+            a = np.sum(self.input_seqs_ohe[0].reshape(sequence_length, len(self.bases)) * theta_bar)
+            a = -a
+
+        elif self.model_type == 'neighbor':
+
+            # compute gauge-fixed, neighbor latent model
+            gf_theta, a = fix_gauge_neighbor_model(sequence_length, len(self.bases), theta)
+
+        elif self.model_type == 'pairwise':
+
+            # compute gauge-fixed pairwise latent model
+            gf_theta, a = fix_gauge_pairwise_model(sequence_length, len(self.bases), theta)
+
+        return gf_theta.ravel() / norm
+
     def noise_model(self,
                     sequences=None,
                     input_range=None,
@@ -864,13 +1030,74 @@ class NoiseAgnosticModel:
 
         noise_model = Model(inputs=noise_model_input, outputs=next_input)
 
-        if input_range is None and sequences is not None:
-            input_range = np.linspace(min(latent_trait), max(latent_trait), 1000)
-            noise_function = noise_model.predict(input_range)
+        # variables required for gauge fixed latent trait parameters theta
+        sequence_length = len(self.x_train[0])
+        # non-gauge fixed theta
+        theta = self.model.layers[1].get_weights()[0]
+        bias = self.model.layers[1].get_weights()[1]
 
-            # if input range not provided, return input range
-            # so ge nonliearity can be plotted against it.
-            return noise_function, input_range, latent_trait
-        else:
-            noise_function = noise_model.predict(input_range)
-            return noise_function
+        # use gauge-fixing method for the appropriate latent model type
+        if self.model_type == 'additive':
+            theta_df = pd.DataFrame(theta.reshape(sequence_length, len(self.bases)), columns=self.bases)
+            # compute mean centered, thus gauge-fixed, additive latent model
+            gf_theta, theta_bar = _center_matrix(theta_df)
+            gf_theta = gf_theta.values
+            # compute offset term which separates latent trait from gauge fixed latent trait
+            a = np.sum(sequences_ohe[0].reshape(sequence_length, len(self.bases)) * theta_bar)
+            a = -a
+
+        elif self.model_type == 'neighbor':
+
+            # compute gauge-fixed, neighbor latent model
+            gf_theta, a = fix_gauge_neighbor_model(sequence_length, len(self.bases), theta)
+
+        elif self.model_type == 'pairwise':
+
+            # compute gauge-fixed pairwise latent model
+            gf_theta, a = fix_gauge_pairwise_model(sequence_length, len(self.bases), theta)
+
+        # this conditional currently outputs the not gauge-fixed, original, GE nonlinearity
+        if gauge_fix is False:
+            if input_range is None and sequences is not None:
+                input_range = np.linspace(min(latent_trait), max(latent_trait), 1000)
+                noise_function = noise_model.predict(input_range)
+
+                # if input range not provided, return input range
+                # so ge nonliearity can be plotted against it.
+                return noise_function, input_range, latent_trait
+            else:
+                noise_function = noise_model.predict(input_range)
+                return noise_function
+
+        # this part of the condition returns the gauge fixed latent trait
+        elif gauge_fix is True:
+
+            # first compute norm which will be used to rescale the parameters theta
+            # compute phi using training sequences
+            phi_for_computing_norm = get_1st_layer_output([self.input_seqs_ohe])
+
+            # compute the norm using phi_for_computing_norm as:
+            norm = np.sqrt(np.var(phi_for_computing_norm[0]))
+
+            # rescale the gf_theta
+            theta_gf_rescaled = gf_theta.ravel() / norm
+
+            # create a copy of the model which will gauge fixed, this
+            # way the user can retain non-gauge-fixed model if they want
+            gauge_fixed_model = tf.keras.models.clone_model(self.model)
+
+            # update the weights of the latent model with the gauge fixed parameters, and compute gauge fixed phi
+            gauge_fixed_model.layers[1].set_weights([theta_gf_rescaled.reshape(len(theta_gf_rescaled), 1),
+                                              self.model.layers[1].get_weights()[1]])
+
+            # update the weights of the latent model with the gauge fixed parameters, and compute gauge fixed phi
+            gauge_fixed_model.layers[1].set_weights([theta_gf_rescaled.reshape(len(theta_gf_rescaled), 1),
+                                              self.model.layers[1].get_weights()[1]])
+
+            get_gauge_fixed_phi = K.function([gauge_fixed_model.layers[0].input], [gauge_fixed_model.layers[1].output])
+            phi_gauge_fixed = get_gauge_fixed_phi([sequences_ohe])
+
+            input_range_gf_phi = np.linspace(min(phi_gauge_fixed[0]), max(phi_gauge_fixed[0]), 1000)
+
+            noise_model = noise_model.predict([(input_range_gf_phi-bias)*norm+a+bias])
+            return noise_model, input_range_gf_phi, phi_gauge_fixed[0]
