@@ -88,6 +88,9 @@ def onehot_encode_array(data, bases_dict, ohe_single_batch_size=10000):
     check(isinstance(ohe_single_batch_size, (int, np.int64)),
           'type(ohe_single_batch_size) = %s must be of type int or numpy.int64' % type(ohe_single_batch_size))
 
+    # check that data[0] passed has length, which will be used to get sequence length
+    check(hasattr(data[0], '__len__'), 'entered sequence data does not have length, needs to have length > 0')
+
     sequence_length = len(data[0])
 
     # container list for batches of oh-encoded sequences
@@ -395,20 +398,36 @@ def na_plots_for_mavenn_demo(loss_history,
     ax[1].set_ylabel('additive parameters')
     ax[1].set_xlabel('position')
 
-    theta_df = pd.DataFrame(NAR.nn_model().layers[1].get_weights()[0].reshape(39, 4),columns=['A', 'C', 'G', 'T'])
+    #theta_df = pd.DataFrame(NAR.nn_model().layers[1].get_weights()[0].reshape(39, 4),columns=['A', 'C', 'G', 'T'])
+    theta_df = pd.DataFrame(NAR.return_theta().reshape(39, 4), columns=['A', 'C', 'G', 'T'])
 
-    additive_logo = logomaker.Logo(theta_df/np.sqrt(np.sum(theta_df.values.ravel()**2)),
+
+    additive_logo = logomaker.Logo(theta_df,
                                    center_values=True,
                                    ax=ax[1])
 
     # view the inferred noise model as a heatmap
-    noise_model_heatmap = sns.heatmap(noise_model.T, cmap='Greens', ax=ax[2])
+    # noise_model_heatmap = sns.heatmap(noise_model.T, cmap='Greens', ax=ax[2])
+    # ax[2].invert_yaxis()
+    # ax[2].set_xticks(([0,int(len(phi_range)/2), len(phi_range)-2]), minor=False)
+    # ax[2].set_xticklabels(([str(phi_range[0]), 0, str(phi_range[len(phi_range)-1])]), minor=False)
+    # ax[2].set_ylabel(' bin numbers')
+    # ax[2].set_xlabel('latent trait ($\phi$)')
+
+    if noise_model.T[noise_model.T.shape[0] - 1][0] > noise_model.T[noise_model.T.shape[0] - 1][
+                noise_model.T.shape[1] - 1]:
+        noise_model_heatmap = sns.heatmap(pd.DataFrame(noise_model.T).loc[::1, ::-1], cmap='Greens', ax=ax[2])
+    else:
+        noise_model_heatmap = sns.heatmap(noise_model.T, cmap='Greens', ax=ax[2])
     ax[2].invert_yaxis()
-    ax[2].set_xticks(([0,int(len(phi_range)/2), len(phi_range)-2]), minor=False)
-    ax[2].set_xticklabels(([str(phi_range[0]), 0, str(phi_range[len(phi_range)-1])]), minor=False)
+    ax[2].set_xticks(([0, int(len(phi_range) / 2), len(phi_range) - 2]), minor=False)
+    middle_tick = str(phi_range[int(len(phi_range) / 2)])
+    ax[2].set_xticklabels(([str(phi_range[0])[1:5], middle_tick[1:5], str(phi_range[len(phi_range) - 1])[1:5]]),
+                       minor=False)
     ax[2].set_ylabel(' bin numbers')
     ax[2].set_xlabel('latent trait ($\phi$)')
 
+
     plt.tight_layout()
     plt.show()
 
@@ -509,6 +528,9 @@ def load_olson_data_GB1():
 def get_example_dataset(name='MPSA'):
     """
 
+    Load example sequence-function datasets that
+    come with the mavenn package.
+
     Parameters:
     -----------
 
@@ -517,6 +539,7 @@ def get_example_dataset(name='MPSA'):
         ('MPSA', 'Sort-Seq', 'GB1-DMS')
 
     Returns:
+    --------
     X, y: (array-like)
         An array containing sequences X and an
         array containing their target values y
@@ -546,3 +569,180 @@ def get_example_dataset(name='MPSA'):
 
         gb1_df = load_olson_data_GB1()
         return gb1_df['sequence'].values, gb1_df['values'].values
+
+
+@handle_errors
+def _center_matrix(df):
+    """
+    Centers each row of a matrix about zero by subtracting out the mean.
+    This method is used for gauge-fixing additive latent models
+
+    parameters
+    ----------
+
+    df: (pd.DataFrame)
+        a (L x C) pandas dataframe which contains additive latent model
+        parameters
+
+    returns
+    -------
+    out_df: (pd.DataFrame)
+        dataframe that is mean centered.
+
+    theta_bar: (array-like)
+        mean of parameters at position l
+
+    """
+
+    # Compute mean value of each row
+    means = df.mean(axis=1).values
+    theta_bar = -df.mean(axis=1).values[:, np.newaxis]
+
+    # Subtract out means
+    out_df = df.copy()
+    out_df.loc[:, :] = df.values - means[:, np.newaxis]
+
+    return out_df, theta_bar
+
+@handle_errors
+def fix_gauge_neighbor_model(sequenceLength,
+                             alphabetSize,
+                             theta):
+    """Calculates the hierarchical gauge that prioritizes the
+    constant feature over neighbor features and the neighbor
+    features' coefficients are in the canonical gauge.
+
+    parameters
+    ----------
+    sequenceLength: (int)
+        length of one input sequence
+
+    alphabetSize: (int)
+        number of unique bases or amino-acids; e.g. 4 for DNA and 20 for proteins
+
+    theta: (array-like)
+        the parameters of the latent neigbhor model
+
+    returns
+    -------
+    thetaGaugeFixed: (array-like)
+        the gauge fixed neighbor parameters
+
+    centeringConstant: (array-like)
+        the additive shift relating phi to phi''
+
+    """
+
+    theta = theta.reshape(sequenceLength - 1, alphabetSize, alphabetSize)
+
+    # center parameter values corresponding to a fixed first index, i.e. fixed neighboring position pair:
+    centeringConstant = sum(theta.mean(axis=(1, 2)))
+    thetaCentered = theta - theta.mean(axis=(1, 2), keepdims=True)
+
+    # calculate canonical gauge for neighbor features:
+    thetaGaugeFixed = thetaCentered
+    if sequenceLength > 2:
+        sRow = np.sum(thetaCentered, axis=2,
+                      keepdims=True)  # define array that broadcasts with theta and contains theta's row sums
+        sCol = np.sum(thetaCentered, axis=1,
+                      keepdims=True)  # define array that broadcasts with theta and contains theta's column sums
+        pos = 0
+        thetaGaugeFixed[pos] = thetaCentered[pos] + 1 / (2 * alphabetSize) * (-sCol[pos] + sRow[pos + 1].transpose())
+        for pos in range(1, sequenceLength - 2):
+            thetaGaugeFixed[pos] = thetaCentered[pos] + 1 / (2 * alphabetSize) * (
+            sCol[pos - 1].transpose() - sRow[pos] - sCol[pos] + sRow[pos + 1].transpose())
+        pos = sequenceLength - 2
+        thetaGaugeFixed[pos] = thetaCentered[pos] + 1 / (2 * alphabetSize) * (sCol[pos - 1].transpose() - sRow[pos])
+    thetaGaugeFixed = thetaGaugeFixed.ravel()
+
+    return thetaGaugeFixed, centeringConstant
+
+
+
+@handle_errors
+def kronecker_delta(n1, n2):
+    """
+
+    helper method implementing Kronecker delta,
+    used in fix_gauge_pairwise_model
+    """
+
+    if n1 == n2:
+        return 1
+    else:
+        return 0
+
+
+@handle_errors
+def fix_gauge_pairwise_model(sequenceLength,
+                             alphabetSize,
+                             theta):
+    """
+    Calculates the hierarchical gauge that prioritizes the
+    constant feature over pairwise features and the pairwise
+    features' coefficients are in the canonical gauge.
+
+     parameters
+    ----------
+    sequenceLength: (int)
+        length of one input sequence
+
+    alphabetSize: (int)
+        number of unique bases or amino-acids; e.g. 4 for DNA and 20 for proteins
+
+    theta: (array-like)
+        the parameters of the latent neigbhor model
+
+    returns
+    -------
+    thetaGaugeFixed: (array-like)
+        the gauge fixed pairwise parameters
+
+    centeringConstant: (array-like)
+        the additive shift relating phi to phi''
+
+     """
+
+    numPosPairs = int(sequenceLength * (sequenceLength - 1) / 2)
+    theta = theta.reshape(numPosPairs, alphabetSize, alphabetSize)
+
+    # center parameter values corresponding to a fixed first index, i.e. fixed position pair:
+    centeringConstant = sum(theta.mean(axis=(1, 2)))
+    thetaCentered = theta - theta.mean(axis=(1, 2), keepdims=True)
+
+    # calculate canonical gauge for pairwise features:
+    thetaGaugeFixed = thetaCentered
+    if sequenceLength > 2:
+        sRow = np.sum(thetaCentered, axis=2,
+                      keepdims=True)  # define array that broadcasts with theta and contains theta's row sums
+        sCol = np.sum(thetaCentered, axis=1,
+                      keepdims=True)  # define array that broadcasts with theta and contains theta's column sums
+
+        # reshape thetaGaugeFixed, sRow, and sCol so that they are inexed as [pos1][pos2,char1,char2] where pos1<pos2
+        l = [i for i in range(1, sequenceLength)]
+        splittingPoints = [sum(l[-k:]) for k in range(1,
+                                                      sequenceLength)]  # [(sequenceLength-1), (sequenceLength-1+sequenceLength-2),...,(sequenceLength-1+sequenceLength-2+...+1)]
+        thetaGaugeFixedReshaped = np.array(np.split(thetaGaugeFixed, splittingPoints, axis=0), dtype=object)
+        sRowReshaped = np.array(np.split(sRow, splittingPoints, axis=0), dtype=object)
+        sColReshaped = np.array(np.split(sCol, splittingPoints, axis=0), dtype=object)
+
+        c = 1 / (alphabetSize * (sequenceLength - 1))
+        for pos1 in range(sequenceLength - 1):
+            for pos2 in range(pos1 + 1, sequenceLength):
+                for pos in range(pos1):
+                    thetaGaugeFixedReshaped[pos1][pos2 - (pos1 + 1)] += c * (-(sequenceLength - 2)) ** kronecker_delta(
+                        pos, pos2) * sColReshaped[pos][pos1 - (pos + 1)].transpose()
+                for pos in range(pos1 + 1, sequenceLength):
+                    thetaGaugeFixedReshaped[pos1][pos2 - (pos1 + 1)] += c * (-(sequenceLength - 2)) ** kronecker_delta(
+                        pos, pos2) * sRowReshaped[pos1][pos - (pos1 + 1)]
+                for pos in range(pos2):
+                    thetaGaugeFixedReshaped[pos1][pos2 - (pos1 + 1)] += c * (-(sequenceLength - 2)) ** kronecker_delta(
+                        pos, pos1) * sColReshaped[pos][pos2 - (pos + 1)]
+                for pos in range(pos2 + 1, sequenceLength):
+                    thetaGaugeFixedReshaped[pos1][pos2 - (pos1 + 1)] += c * (-(sequenceLength - 2)) ** kronecker_delta(
+                        pos, pos1) * sRowReshaped[pos2][pos - (pos2 + 1)].transpose()
+    thetaGaugeFixed = thetaGaugeFixed.ravel()
+
+    return thetaGaugeFixed, centeringConstant
+
+
